@@ -12,6 +12,7 @@ macro_rules! impl_from_slice_into_smartptr {
         mutability: $mut:ident,
     ) => {
         #[cfg(feature = "alloc")]
+        #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
         impl<S: crate::spec::Spec> From<&$ty<S>> for $($smartptr)::* <$ty<S>> {
             fn from(s: &$ty<S>) -> Self {
                 let inner: &str = s.as_str();
@@ -136,12 +137,13 @@ macro_rules! impl_cmp2_as_str {
 /// * type conversion
 ///     + `AsRef<&str> for $ty`
 ///     + `AsRef<&$ty> for $ty`
+///     + `From<&$ty>` for Cow<$ty>`
 ///     + `From<&$ty>` for Arc<$ty>`
 ///     + `From<&$ty>` for Box<$ty>`
 ///     + `From<&$ty>` for Rc<$ty>`
 ///     + `From<&$ty> for &str`
 ///     + `TryFrom<&str> for &$ty`
-///     + `AsRef<$ty<IriSpec>> for $ty<UriSpec>`
+///     + `TryFrom<&[u8]> for &$ty`
 /// * comparison (only `PartialEq` impls are listed, but `PartialOrd` is also implemented).
 ///     + `PartialEq<$ty> for $ty`
 ///     + `str` and `$ty`
@@ -195,6 +197,7 @@ macro_rules! define_custom_string_slice {
             /// # Safety
             ///
             /// The given string must be valid.
+            #[must_use]
             pub(crate) unsafe fn new_maybe_unchecked(s: &str) -> &Self {
                 debug_assert_eq!($validate::<S>(s), Ok(()));
                 // It is caller's responsibility to guarantee the given string is valid.
@@ -212,24 +215,28 @@ macro_rules! define_custom_string_slice {
             ///
             /// The given string must be valid.
             #[inline]
+            #[must_use]
             unsafe fn new_always_unchecked(s: &str) -> &Self {
                 &*(s as *const str as *const Self)
             }
 
             /// Returns `&str`.
             #[inline]
+            #[must_use]
             pub fn as_str(&self) -> &str {
                 self.as_ref()
             }
 
             /// Returns the string length.
             #[inline]
+            #[must_use]
             pub fn len(&self) -> usize {
                 self.as_str().len()
             }
 
             /// Returns whether the string is empty.
             #[inline]
+            #[must_use]
             pub fn is_empty(&self) -> bool {
                 self.as_str().is_empty()
             }
@@ -285,6 +292,14 @@ macro_rules! define_custom_string_slice {
             }
         }
 
+        #[cfg(feature = "alloc")]
+        impl<'a, S: crate::spec::Spec> From<&'a $ty<S>> for alloc::borrow::Cow<'a, $ty<S>> {
+            #[inline]
+            fn from(s: &'a $ty<S>) -> Self {
+                alloc::borrow::Cow::Borrowed(s)
+            }
+        }
+
         impl_from_slice_into_smartptr! {
             ty: $ty,
             smartptr: alloc::sync::Arc,
@@ -322,6 +337,19 @@ macro_rules! define_custom_string_slice {
             }
         }
 
+        impl<'a, S: crate::spec::Spec> core::convert::TryFrom<&'a [u8]> for &'a $ty<S> {
+            type Error = crate::validate::Error;
+
+            #[inline]
+            fn try_from(bytes: &'a [u8]) -> Result<Self, Self::Error> {
+                let s = core::str::from_utf8(bytes).map_err(|_| crate::validate::Error::new())?;
+                match $validate::<S>(s) {
+                    Ok(()) => Ok(unsafe { $ty::new_always_unchecked(s) }),
+                    Err(e) => Err(e),
+                }
+            }
+        }
+
         impl_cmp!(str, str, $ty<S>);
         impl_cmp!(str, &str, $ty<S>);
         impl_cmp!(str, str, &$ty<S>);
@@ -334,15 +362,6 @@ macro_rules! define_custom_string_slice {
             }
         }
 
-        impl AsRef<$ty<crate::spec::IriSpec>> for $ty<crate::spec::UriSpec> {
-            fn as_ref(&self) -> &$ty<crate::spec::IriSpec> {
-                unsafe {
-                    // This is safe because URIs are subset of IRIs.
-                    <$ty<crate::spec::IriSpec>>::new_maybe_unchecked(self.as_str())
-                }
-            }
-        }
-
         /// Serde deserializer implementation.
         #[cfg(feature = "serde")]
         mod __serde_slice {
@@ -378,6 +397,7 @@ macro_rules! define_custom_string_slice {
 
             // About `'de` and `'a`, see
             // <https://serde.rs/lifetimes.html#the-deserializede-lifetime>.
+            #[cfg_attr(docsrs, doc(cfg(feature = "serde")))]
             impl<'de: 'a, 'a, S: 'de + crate::spec::Spec> Deserialize<'de> for &'a $ty<S> {
                 #[inline]
                 fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
@@ -415,11 +435,13 @@ macro_rules! define_custom_string_slice {
 ///     + `ToOwned<Owned = $ty> for $slice`
 ///     + `From<&$slice> for $ty`
 ///     + `From<$ty> for String`
+///     + `From<$ty> for Cow<'_, $slice>`
+///     + `From<$ty> for Box<$slice>`
 ///     + `TryFrom<&str> for $ty`
+///     + `TryFrom<&[u8]> for $ty`
 ///     + `TryFrom<String> for $ty`
 ///     + `FromStr for $ty`
 ///     + `Deref<Target = $slice> for $ty`
-///     + `AsRef<$slice<IriSpec>> for $ty<UriSpec>`
 /// * comparison (only `PartialEq` impls are listed, but `PartialOrd` is also implemented.
 ///     + `PartialEq<$ty> for $ty`
 ///     + `$slice` and `str`
@@ -454,6 +476,11 @@ macro_rules! define_custom_string_slice {
 /// * serde
 ///     + `serde::Serialize`
 ///     + `serde::Deserialize`
+// Note that `From<$ty> for {Arc,Rc}<$slice>` is currently not implemented since
+// this won't reuse allocated memory and hides internal memory reallocation. See
+// <https://github.com/lo48576/iri-string/issues/20#issuecomment-1105207849>.
+// However, this is not decided with firm belief or opinion, so there would be
+// a chance that they are implemented in future.
 #[cfg(feature = "alloc")]
 macro_rules! define_custom_string_owned {
     (
@@ -467,12 +494,13 @@ macro_rules! define_custom_string_owned {
         $(#[$meta])*
         // `#[derive(..)]` cannot be used here, because it adds `S: DerivedTrait` bounds automatically.
         #[cfg(feature = "alloc")]
-        #[cfg_attr(feature = "serde", derive(serde::Serialize))]
-        #[cfg_attr(feature = "serde", serde(bound = "S: crate::spec::Spec"))]
-        #[cfg_attr(feature = "serde", serde(transparent))]
+        #[cfg_attr(feature = "serde-alloc", derive(serde::Serialize))]
+        #[cfg_attr(feature = "serde-alloc", serde(bound = "S: crate::spec::Spec"))]
+        #[cfg_attr(feature = "serde-alloc", serde(transparent))]
+        #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
         pub struct $ty<S> {
             /// Spec.
-            #[cfg_attr(feature = "serde", serde(skip))]
+            #[cfg_attr(feature = "serde-alloc", serde(skip))]
             _spec: core::marker::PhantomData<fn() -> S>,
             /// Inner data.
             inner: alloc::string::String,
@@ -489,6 +517,7 @@ macro_rules! define_custom_string_owned {
             ///
             /// The given string must be valid.
             #[inline]
+            #[must_use]
             pub(crate) unsafe fn new_always_unchecked(s: alloc::string::String) -> Self {
                 Self {
                     _spec: core::marker::PhantomData,
@@ -503,16 +532,49 @@ macro_rules! define_custom_string_owned {
             /// # Safety
             ///
             /// The given string must be valid.
+            #[must_use]
             pub(crate) unsafe fn new_maybe_unchecked(s: alloc::string::String) -> Self {
                 debug_assert_eq!($validate::<S>(&s), Ok(()));
                 Self::new_always_unchecked(s)
             }
 
+            /// Returns a mutable reference to the inner string buffer.
+            ///
+            /// This may be useful to implement inline modification algorithm,
+            /// but be careful as this method itself cannot validate the new
+            /// content.
+            ///
+            /// # Safety
+            ///
+            /// The content after modification must be valid.
+            #[inline]
+            #[must_use]
+            // TODO: Use wrapper type to enforce validation on finish?
+            pub(crate) fn as_inner_mut(&mut self) -> &mut alloc::string::String {
+                &mut self.inner
+            }
+
             /// Shrinks the capacity of the inner buffer to match its length.
             #[inline]
             pub fn shrink_to_fit(&mut self) {
                 self.inner.shrink_to_fit()
             }
+
+            /// Returns the internal buffer capacity in bytes.
+            #[inline]
+            #[must_use]
+            pub fn capacity(&self) -> usize {
+                self.inner.capacity()
+            }
+
+            /// Returns the borrowed IRI string slice.
+            ///
+            /// This is equivalent to `&*self`.
+            #[inline]
+            #[must_use]
+            pub fn as_slice(&self) -> &$slice<S> {
+                self.as_ref()
+            }
         }
 
         impl<S: crate::spec::Spec> core::fmt::Debug for $ty<S> {
@@ -620,6 +682,25 @@ macro_rules! define_custom_string_owned {
             }
         }
 
+        impl<'a, S: crate::spec::Spec> From<$ty<S>> for alloc::borrow::Cow<'a, $slice<S>> {
+            #[inline]
+            fn from(s: $ty<S>) -> alloc::borrow::Cow<'a, $slice<S>> {
+                alloc::borrow::Cow::Owned(s)
+            }
+        }
+
+        impl<S: crate::spec::Spec> From<$ty<S>> for alloc::boxed::Box<$slice<S>> {
+            #[inline]
+            fn from(s: $ty<S>) -> alloc::boxed::Box<$slice<S>> {
+                let inner: alloc::string::String = s.into();
+                let buf = alloc::boxed::Box::<str>::from(inner);
+                unsafe {
+                    let raw: *mut str = alloc::boxed::Box::into_raw(buf);
+                    alloc::boxed::Box::<$slice<S>>::from_raw(raw as *mut $slice<S>)
+                }
+            }
+        }
+
         impl<S: crate::spec::Spec> core::convert::TryFrom<&'_ str> for $ty<S> {
             type Error = crate::validate::Error;
 
@@ -629,6 +710,16 @@ macro_rules! define_custom_string_owned {
             }
         }
 
+        impl<S: crate::spec::Spec> core::convert::TryFrom<&'_ [u8]> for $ty<S> {
+            type Error = crate::validate::Error;
+
+            #[inline]
+            fn try_from(bytes: &[u8]) -> Result<Self, Self::Error> {
+                let s = core::str::from_utf8(bytes).map_err(|_| crate::validate::Error::new())?;
+                <&$slice<S>>::try_from(s).map(Into::into)
+            }
+        }
+
         impl<S: crate::spec::Spec> core::convert::TryFrom<alloc::string::String> for $ty<S> {
             type Error = crate::types::CreationError<alloc::string::String>;
 
@@ -666,12 +757,6 @@ macro_rules! define_custom_string_owned {
             }
         }
 
-        impl AsRef<$slice<crate::spec::IriSpec>> for $ty<crate::spec::UriSpec> {
-            fn as_ref(&self) -> &$slice<crate::spec::IriSpec> {
-                AsRef::<$slice<crate::spec::UriSpec>>::as_ref(self).as_ref()
-            }
-        }
-
         impl_cmp!(str, $slice<S>, alloc::borrow::Cow<'_, str>);
         impl_cmp!(str, &$slice<S>, alloc::borrow::Cow<'_, str>);
         impl_cmp2_as_str!(&$slice<S>, alloc::borrow::Cow<'_, $slice<T>>);
@@ -694,11 +779,12 @@ macro_rules! define_custom_string_owned {
         /// Serde deserializer implementation.
         #[cfg(all(feature = "alloc", feature = "serde"))]
         mod __serde_owned {
-            use super::{$slice, $ty};
+            use super::$ty;
 
             use core::{convert::TryFrom, fmt, marker::PhantomData};
 
-            use alloc::{borrow::ToOwned, string::String};
+            #[cfg(feature = "serde-alloc")]
+            use alloc::string::String;
 
             use serde::{
                 de::{self, Visitor},
@@ -722,11 +808,10 @@ macro_rules! define_custom_string_owned {
                 where
                     E: de::Error,
                 {
-                    <&$slice<S>>::try_from(v)
-                        .map(ToOwned::to_owned)
-                        .map_err(E::custom)
+                    <$ty<S> as TryFrom<&str>>::try_from(v).map_err(E::custom)
                 }
 
+                #[cfg(feature = "serde-alloc")]
                 #[inline]
                 fn visit_string<E>(self, v: String) -> Result<Self::Value, E>
                 where
@@ -736,6 +821,7 @@ macro_rules! define_custom_string_owned {
                 }
             }
 
+            #[cfg_attr(docsrs, doc(cfg(feature = "serde")))]
             impl<'de, S: crate::spec::Spec> Deserialize<'de> for $ty<S> {
                 #[inline]
                 fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
@@ -749,7 +835,7 @@ macro_rules! define_custom_string_owned {
     };
 }
 
-/// Implements infallible conversions and other useful traits between two IRI types.
+/// Implements trivial conversions and other useful traits between two IRI types.
 ///
 /// Implemented traits:
 ///
@@ -793,7 +879,7 @@ macro_rules! define_custom_string_owned {
 ///     + `$from_owned` and `$to_owned`
 ///         - `PartialEq<$from_owned> for $to_owned`
 ///         - `PartialEq<$to_owned> for $from_owned`
-macro_rules! impl_infallible_conv_between_iri {
+macro_rules! impl_trivial_conv_between_iri {
     (
         from_slice: $from_slice:ident,
         from_owned: $from_owned:ident,
