@@ -3,14 +3,19 @@
 #[cfg(feature = "alloc")]
 use alloc::string::String;
 
+use crate::components::AuthorityComponents;
+#[cfg(feature = "alloc")]
+use crate::normalize::{Error, NormalizationTask};
+use crate::parser::trusted as trusted_parser;
+#[cfg(feature = "alloc")]
+use crate::raw;
+use crate::spec::Spec;
+#[cfg(feature = "alloc")]
+use crate::task::{Error as TaskError, ProcessAndWrite};
+use crate::types::{RiAbsoluteStr, RiFragmentStr, RiReferenceStr};
 #[cfg(feature = "alloc")]
 use crate::types::{RiAbsoluteString, RiFragmentString, RiReferenceString};
-use crate::{
-    raw,
-    spec::Spec,
-    types::{RiAbsoluteStr, RiFragmentStr, RiReferenceStr},
-    validate::iri,
-};
+use crate::validate::iri;
 
 define_custom_string_slice! {
     /// A borrowed string of an absolute IRI possibly with fragment part.
@@ -143,8 +148,9 @@ impl<S: Spec> RiStr<S> {
     /// assert_eq!(fragment, None);
     /// # Ok::<_, Error>(())
     /// ```
+    #[must_use]
     pub fn to_absolute_and_fragment(&self) -> (&RiAbsoluteStr<S>, Option<&RiFragmentStr<S>>) {
-        let (prefix, fragment) = raw::split_fragment(self.as_str());
+        let (prefix, fragment) = trusted_parser::split_fragment(self.as_str());
         let prefix = unsafe {
             // This is safe because the an IRI without fragment part is also an absolute IRI.
             RiAbsoluteStr::new_maybe_unchecked(prefix)
@@ -177,8 +183,9 @@ impl<S: Spec> RiStr<S> {
     /// ```
     ///
     /// [`RiAbsoluteStr`]: struct.RiAbsoluteStr.html
+    #[must_use]
     pub fn to_absolute(&self) -> &RiAbsoluteStr<S> {
-        let prefix_len = raw::split_fragment(self.as_str()).0.len();
+        let prefix_len = trusted_parser::split_fragment(self.as_str()).0.len();
         unsafe {
             // This is safe because an IRI without the fragment part (and a leading `#` character)
             // is also an absolute IRI.
@@ -186,6 +193,172 @@ impl<S: Spec> RiStr<S> {
         }
     }
 
+    /// Returns `true` if the IRI is already normalized.
+    ///
+    /// This returns the same result as
+    /// `self.normalize.map_or(false, |normalized| normalized == self))`, but
+    /// does this more efficiently without heap allocation.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # #[derive(Debug)] struct Error;
+    /// # impl From<iri_string::validate::Error> for Error {
+    /// #     fn from(e: iri_string::validate::Error) -> Self { Self } }
+    /// # impl<T> From<iri_string::task::Error<T>> for Error {
+    /// #     fn from(e: iri_string::task::Error<T>) -> Self { Self } }
+    /// # #[cfg(feature = "alloc")] {
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("HTTP://example.COM/foo/./bar/%2e%2e/../baz?query#fragment")?;
+    /// assert!(!iri.is_normalized());
+    ///
+    /// let normalized = iri.normalize()?;
+    /// assert_eq!(normalized, "http://example.com/baz?query#fragment");
+    /// assert!(normalized.is_normalized());
+    /// # }
+    /// # Ok::<_, Error>(())
+    /// ```
+    #[must_use]
+    #[inline]
+    pub fn is_normalized(&self) -> bool {
+        trusted_parser::is_normalized::<S>(self.as_str())
+    }
+
+    /// Returns the normalized IRI.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # #[derive(Debug)] struct Error;
+    /// # impl From<iri_string::validate::Error> for Error {
+    /// #     fn from(e: iri_string::validate::Error) -> Self { Self } }
+    /// # impl<T> From<iri_string::task::Error<T>> for Error {
+    /// #     fn from(e: iri_string::task::Error<T>) -> Self { Self } }
+    /// # #[cfg(feature = "alloc")] {
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("HTTP://example.COM/foo/./bar/%2e%2e/../baz?query#fragment")?;
+    ///
+    /// let normalized = iri.normalize()?;
+    /// assert_eq!(normalized, "http://example.com/baz?query#fragment");
+    /// # }
+    /// # Ok::<_, Error>(())
+    /// ```
+    #[cfg(feature = "alloc")]
+    #[inline]
+    pub fn normalize(&self) -> Result<RiString<S>, TaskError<Error>> {
+        NormalizationTask::from(self).allocate_and_write()
+    }
+}
+
+/// Components getters.
+impl<S: Spec> RiStr<S> {
+    /// Returns the scheme.
+    ///
+    /// The following colon is truncated.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use iri_string::validate::Error;
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("http://example.com/pathpath?queryquery#fragfrag")?;
+    /// assert_eq!(iri.scheme_str(), "http");
+    /// # Ok::<_, Error>(())
+    /// ```
+    #[inline]
+    #[must_use]
+    pub fn scheme_str(&self) -> &str {
+        trusted_parser::extract_scheme_absolute(self.as_str())
+    }
+
+    /// Returns the authority.
+    ///
+    /// The leading `//` is truncated.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use iri_string::validate::Error;
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("http://example.com/pathpath?queryquery#fragfrag")?;
+    /// assert_eq!(iri.authority_str(), Some("example.com"));
+    /// # Ok::<_, Error>(())
+    /// ```
+    ///
+    /// ```
+    /// # use iri_string::validate::Error;
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("urn:uuid:10db315b-fcd1-4428-aca8-15babc9a2da2")?;
+    /// assert_eq!(iri.authority_str(), None);
+    /// # Ok::<_, Error>(())
+    /// ```
+    #[inline]
+    #[must_use]
+    pub fn authority_str(&self) -> Option<&str> {
+        trusted_parser::extract_authority_absolute(self.as_str())
+    }
+
+    /// Returns the path.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use iri_string::validate::Error;
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("http://example.com/pathpath?queryquery#fragfrag")?;
+    /// assert_eq!(iri.path_str(), "/pathpath");
+    /// # Ok::<_, Error>(())
+    /// ```
+    ///
+    /// ```
+    /// # use iri_string::validate::Error;
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("urn:uuid:10db315b-fcd1-4428-aca8-15babc9a2da2")?;
+    /// assert_eq!(iri.path_str(), "uuid:10db315b-fcd1-4428-aca8-15babc9a2da2");
+    /// # Ok::<_, Error>(())
+    /// ```
+    #[inline]
+    #[must_use]
+    pub fn path_str(&self) -> &str {
+        trusted_parser::extract_path_absolute(self.as_str())
+    }
+
+    /// Returns the query.
+    ///
+    /// The leading question mark (`?`) is truncated.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use iri_string::validate::Error;
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("http://example.com/pathpath?queryquery#fragfrag")?;
+    /// assert_eq!(iri.query_str(), Some("queryquery"));
+    /// # Ok::<_, Error>(())
+    /// ```
+    ///
+    /// ```
+    /// # use iri_string::validate::Error;
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("urn:uuid:10db315b-fcd1-4428-aca8-15babc9a2da2")?;
+    /// assert_eq!(iri.query_str(), None);
+    /// # Ok::<_, Error>(())
+    /// ```
+    #[inline]
+    #[must_use]
+    pub fn query_str(&self) -> Option<&str> {
+        trusted_parser::extract_query(self.as_str())
+    }
+
     /// Returns the fragment part if exists.
     ///
     /// A leading `#` character is truncated if the fragment part exists.
@@ -215,9 +388,41 @@ impl<S: Spec> RiStr<S> {
     /// # Ok::<_, Error>(())
     /// ```
     #[inline]
+    #[must_use]
     pub fn fragment(&self) -> Option<&RiFragmentStr<S>> {
         AsRef::<RiReferenceStr<S>>::as_ref(self).fragment()
     }
+
+    /// Returns the authority components.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use iri_string::validate::Error;
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("http://user:pass@example.com:8080/pathpath?queryquery")?;
+    /// let authority = iri.authority_components()
+    ///     .expect("authority is available");
+    /// assert_eq!(authority.userinfo(), Some("user:pass"));
+    /// assert_eq!(authority.host(), "example.com");
+    /// assert_eq!(authority.port(), Some("8080"));
+    /// # Ok::<_, Error>(())
+    /// ```
+    ///
+    /// ```
+    /// # use iri_string::validate::Error;
+    /// use iri_string::types::IriStr;
+    ///
+    /// let iri = IriStr::new("urn:uuid:10db315b-fcd1-4428-aca8-15babc9a2da2")?;
+    /// assert_eq!(iri.authority_str(), None);
+    /// # Ok::<_, Error>(())
+    /// ```
+    #[inline]
+    #[must_use]
+    pub fn authority_components(&self) -> Option<AuthorityComponents<'_>> {
+        AuthorityComponents::from_iri(self.as_ref())
+    }
 }
 
 #[cfg(feature = "alloc")]
@@ -262,6 +467,7 @@ impl<S: Spec> RiString<S> {
     /// assert_eq!(fragment, None);
     /// # Ok::<_, Error>(())
     /// ```
+    #[must_use]
     pub fn into_absolute_and_fragment(self) -> (RiAbsoluteString<S>, Option<RiFragmentString<S>>) {
         let (prefix, fragment) = raw::split_fragment_owned(self.into());
         let prefix = unsafe {
@@ -296,6 +502,7 @@ impl<S: Spec> RiString<S> {
     /// ```
     ///
     /// [`RiAbsoluteString`]: struct.RiAbsoluteString.html
+    #[must_use]
     pub fn into_absolute(self) -> RiAbsoluteString<S> {
         let mut s: String = self.into();
         raw::remove_fragment(&mut s);
@@ -314,7 +521,7 @@ impl<S: Spec> RiString<S> {
     }
 }
 
-impl_infallible_conv_between_iri! {
+impl_trivial_conv_between_iri! {
     from_slice: RiStr,
     from_owned: RiString,
     to_slice: RiReferenceStr,
